@@ -1,6 +1,9 @@
+from pythonql.algebra.operator import plan_from_list
+from pythonql.algebra.operators import *
 from pythonql.PQTuple import PQTuple
 from pythonql.helpers import flatten
 from pythonql.Rewriter import rewrite
+from pythonql.debug import Debug
 import json
 import types
 
@@ -113,12 +116,19 @@ def emptyTuple(schema):
   return PQTuple([None] * len(schema), schema)
 
 # Execute the query
-def PyQuery( clauses, prior_locs, returnType ):
+def PyQuery( clauses, prior_locs, prior_globs, returnType ):
   data = []
   data.append( emptyTuple([]) )
-  clauses = rewrite(clauses, prior_locs)
-  for c in clauses:
-    data = processClause(c, data, prior_locs)
+
+  clauses = list(clauses)
+  clauses.reverse()
+
+  plan = plan_from_list(clauses)
+  plan = rewrite(plan, prior_locs)
+
+  if Debug().print_optimized:
+      print("Rewritten query:",plan)
+  data = plan.execute(data, prior_locs, prior_globs)
   if returnType == "gen":
     return data
   elif returnType == "list":
@@ -128,71 +138,44 @@ def PyQuery( clauses, prior_locs, returnType ):
   else:
     return dict(data)
 
-# Process clauses
-def processClause(c, table, prior_locs):
-  if c["name"] == "select":
-    return processSelectClause(c, table, prior_locs)
-  elif c["name"] == "for":
-    return processForClause(c, table, prior_locs)
-  elif c["name"] == "let":
-    return processLetClause(c, table, prior_locs)
-  elif c["name"] == "join":
-    return processJoin(c, table, prior_locs)
-  elif c["name"] == "db_source":
-    return c["database"].execute(c["query"],c["tuple_vars"],c["vars"])
-  elif c["name"] == "match":
-    return processMatchClause(c, table, prior_locs)
-  elif c["name"] == "count":
-    return processCountClause(c, table, prior_locs)
-  elif c["name"] == "where":
-    return processWhereClause(c, table, prior_locs)
-  elif c["name"] == "groupby":
-    return processGroupByClause(c, table, prior_locs)
-  elif c["name"] == "orderby":
-    return processOrderByClause(c, table, prior_locs)
-  elif c["name"] == "window":
-    return processWindowClause(c, table, prior_locs)
-  else:
-    raise Exception("Unknown clause %s encountered" % c["name"] )
-  
 # Process Select clause
 # We still keep that feature of generating tuples for now
-def processSelectClause(c, table, prior_lcs):
+def processSelectClause(c, table, prior_lcs, prior_globs):
   # If this is a list/set comprehension:
-  if "expr" in c:
+  if c.expr:
     # Compile the expression:
-    e = compile(c["expr"].lstrip(), '<string>','eval')
+    e = compile(c.expr.lstrip(), '<string>','eval')
     for t in table:
       lcs = prior_lcs
       lcs.update(t.getDict())
-      yield eval(e,globals(),lcs)
+      yield eval(e,prior_globs,lcs)
 
   else:
-    k_expr = compile(c["key"].lstrip(),'<string>','eval')
-    v_expr = compile(c["value"].lstrip(),'<string>','eval')
+    k_expr = compile(c.key_expr.lstrip(),'<string>','eval')
+    v_expr = compile(c.value_expr.lstrip(),'<string>','eval')
     for t in table:
       lcs = prior_lcs
       lcs.update(t.getDict())
-      k = eval(k_expr,globals(),lcs)
-      v = eval(v_expr,globals(),lcs)
+      k = eval(k_expr,prior_globs,lcs)
+      v = eval(v_expr,prior_globs,lcs)
       yield (k,v)
 
 # Process the for clause. This clause creates a cartesian
 # product of the input table with new sequence
-def processForClause(c, table, prior_lcs):
+def processForClause(c, table, prior_lcs, prior_globs):
   new_schema = None
-  comp_expr = compile(c["expr"].lstrip(), "<string>", "eval")
+  comp_expr = compile(c.expr.lstrip(), "<string>", "eval")
 
   for t in table:
     if not new_schema:
       new_schema = dict(t.schema)
-      for (i,v) in enumerate(c["vars"]):
+      for (i,v) in enumerate(c.vars):
         new_schema[v] = len(t.schema) + i
 
     lcs = prior_lcs
     lcs.update(t.getDict())
-    vals = eval(comp_expr, globals(), lcs)
-    if len(c["vars"]) == 1:
+    vals = eval(comp_expr, prior_globs, lcs)
+    if len(c.vars) == 1:
       for v in vals:
         new_t_data = list(t.tuple)
         new_t_data.append(v)
@@ -202,8 +185,8 @@ def processForClause(c, table, prior_lcs):
     else:
       for v in vals:
         unpack_expr = "[ %s for %s in [ __v ]]" % (
-                      '(' + ",".join(c["vars"]) + ')', c["unpack"])
-        unpacked_vals = eval(unpack_expr, globals(), {'__v':v})
+                      '(' + ",".join(c.vars) + ')', c.unpack)
+        unpacked_vals = eval(unpack_expr, prior_globs, {'__v':v})
         new_t_data = list(t.tuple)
         for tv in unpacked_vals[0]:
           new_t_data.append(tv)
@@ -212,28 +195,28 @@ def processForClause(c, table, prior_lcs):
 
 # Process the let clause. Here we just add a variable to each
 # input tuple
-def processLetClause(c, table, prior_lcs):
-  comp_expr = compile(c["expr"].lstrip(), "<string>", "eval")
+def processLetClause(c, table, prior_lcs, prior_globs):
+  comp_expr = compile(c.expr.lstrip(), "<string>", "eval")
   new_schema = None
   for t in table:
 
     if not new_schema:
       new_schema = dict(t.schema)
-      for (i,v) in enumerate(c["vars"]):
+      for (i,v) in enumerate(c.vars):
         new_schema[v] = len(t.schema) + i
 
     lcs = prior_lcs
     lcs.update(t.getDict())
-    v = eval(comp_expr, globals(), lcs)
-    if len(c["vars"]) == 1:
+    v = eval(comp_expr, prior_globs, lcs)
+    if len(c.vars) == 1:
       t.tuple.append(v)
       new_t = PQTuple( t.tuple, new_schema )
       yield new_t
 
     else:
       unpack_expr = "[ %s for %s in [ __v ]]" % (
-                      '(' + ",".join(c["vars"]) + ')', c["unpack"]) 
-      unpacked_vals = eval(unpack_expr, globals(), {'__v':v})
+                      '(' + ",".join(c.vars) + ')', c.unpack) 
+      unpacked_vals = eval(unpack_expr, prior_globs, {'__v':v})
       new_t_data = list(t.tuple)
       for tv in unpacked_vals[0]:
         new_t_data.append(tv)
@@ -241,18 +224,16 @@ def processLetClause(c, table, prior_lcs):
       yield new_t
 
 # Process a join
-def processJoin(c, table, prior_lcs):
+def processJoin(c, table, prior_lcs, prior_globs, left_arg, right_arg):
   new_schema = None
-  left_arg = c['left']
-  right_arg = c['right']
-  left_conds = c['left_conds']
-  right_conds = c['right_conds']
+  left_conds = c.left_conds
+  right_conds = c.right_conds
  
   join_type = 'nl'
   dir = 'right'
-  if 'hint' in c:
-    join_type = c['hint']['join_type']
-    dir = c['hint']['dir']
+  if c.hint:
+    join_type = c.hint['join_type']
+    dir = c.hint['dir']
 
   if dir == 'left':
     left_arg,right_arg = right_arg,left_arg
@@ -266,18 +247,14 @@ def processJoin(c, table, prior_lcs):
   if join_type == 'index':
     index = {}
     r_data = r_init_data
-    if isinstance(c['right'], list):
-        for c2 in c['right']:
-            r_data = processClause(c2, r_data, prior_lcs)
-    else:
-        r_data = processJoin(c['right'], r_data, prior_lcs)
+    r_data = right_arg.execute(r_data, prior_lcs)
 
     for t in r_data:
         index_tuple = [] 
         for rcond in right_conds:
             lcs = prior_lcs
             lcs.update(t.getDict())
-            rcond_val = eval(rcond, globals(), lcs)
+            rcond_val = eval(rcond, prior_globs, lcs)
             index_tuple.append( rcond_val )
         index_tuple = tuple(index_tuple)
         if not index_tuple in index:
@@ -287,18 +264,13 @@ def processJoin(c, table, prior_lcs):
   # Iterate over the tuples of the left relation and
   # compute the tuple of condition vars
 
-  if isinstance(c['left'], list):
-      for c2 in c['left']:
-          table = processClause(c2, table, prior_lcs)
-  else:
-      table = processJoin(c['left'], table, prior_lcs)
-
+  table = left_arg.execute(table, prior_lcs, prior_globs)
   for t in table:
      cond_tuple = []
      for lcond in left_conds:
          lcs = prior_lcs
          lcs.update(t.getDict())
-         lcond_val = eval(lcond, globals(), lcs)
+         lcond_val = eval(lcond, prior_globs, lcs)
          cond_tuple.append( lcond_val )
      cond_tuple = tuple(cond_tuple)
 
@@ -318,18 +290,14 @@ def processJoin(c, table, prior_lcs):
              continue
      else:
           r_data = r_init_data
-          if isinstance(c['right'], list):
-              for c2 in c['right']:
-                  r_data = processClause(c2, r_data, prior_lcs)
-          else:
-              r_data = processJoin(c['right'], r_data, prior_lcs)
+          r_data = right_arg.execute(r_data, prior_lcs, prior_globs)
 
           for t2 in r_data:
               rcond_tuple = []
               for rcond in right_conds:
                   lcs = prior_lcs
                   lcs.update(t2.getDict())
-                  rcond_val = eval(rcond, globals(), lcs)
+                  rcond_val = eval(rcond, prior_globs, lcs)
                   rcond_tuple.append( rcond_val )
 
               rcond_tuple = tuple(rcond_tuple)
@@ -345,13 +313,13 @@ def processJoin(c, table, prior_lcs):
                  yield new_t
 
 # Process the match claise
-def processMatchClause(c, table, prior_lcs):
-  clause_expr = compile(c['expr'], "<string>", "eval")
+def processMatchClause(c, table, prior_lcs, prior_globs):
+  clause_expr = compile(c.expr, "<string>", "eval")
 
   # Fetch and compile all expressions in the
   # pattern match clause
   e_patterns = []
-  patterns = list(c['pattern'])
+  patterns = list(c.pattern)
   while patterns:
     p = patterns.pop() 
     if 'expr_cond' in p:
@@ -366,24 +334,24 @@ def processMatchClause(c, table, prior_lcs):
   for t in table:
     if not new_schema:
       new_schema = dict(t.schema)
-      for (i,v) in enumerate(c["vars"]):
+      for (i,v) in enumerate(c.vars):
         new_schema[v] = len(t.schema) + i
 
     lcs = prior_lcs
     lcs.update(t.getDict())
-    vals = eval(clause_expr, globals(), lcs)
+    vals = eval(clause_expr, prior_globs, lcs)
 
     for v in vals:
       if not hasattr(v, '__contains__'):
         continue
       
-      new_t_data = list(t.tuple) + [None]*len(c['vars'])
+      new_t_data = list(t.tuple) + [None]*len(c.vars)
       new_t = PQTuple(new_t_data, new_schema)
 
-      if match_pattern(c['pattern'], c['exact'], v, new_t, lcs):
+      if match_pattern(c.pattern, c.exact, v, new_t, lcs, prior_globs):
         yield new_t
 
-def match_pattern(ps, isExact, v, new_t, lcs):
+def match_pattern(ps, isExact, v, new_t, lcs, prior_globs):
   all_heads = []
   for p in [x for x in ps if 'match' in x]:
     match = p['match'][1:-1]
@@ -401,12 +369,12 @@ def match_pattern(ps, isExact, v, new_t, lcs):
       lcs.update({p['bind_to']:v[match]})
 
     if 'expr_cond' in p:
-      val = eval(p['expr_cond'], globals(), lcs)
+      val = eval(p['expr_cond'], prior_globs, lcs)
       if not val:
         return False
       
     if 'pattern' in p:
-      if not match_pattern(p['pattern'], isExact, v[match], new_t, lcs):
+      if not match_pattern(p['pattern'], isExact, v[match], new_t, lcs, prior_globs):
         return False
 
   if isExact and any([x for x in v if x not in all_heads]):
@@ -420,23 +388,23 @@ def match_pattern(ps, isExact, v, new_t, lcs):
   return True
   
 # Process the count clause. Similar to let, but simpler
-def processCountClause(c, table, prior_lcs):
+def processCountClause(c, table, prior_lcs, prior_globs):
   new_schema = None
   for (i,t) in enumerate(table):
 
     if not new_schema:
       new_schema = dict(t.schema)
-      new_schema[c["var"]] = len(t.schema)
+      new_schema[c.var] = len(t.schema)
 
     new_t = PQTuple( t.tuple + [i], new_schema )
     yield new_t
 
 # Process the group-by
-def processGroupByClause(c, table, prior_lcs):
+def processGroupByClause(c, table, prior_lcs, prior_globs):
   gby_aliases = [g if isinstance(g,str) else g[1]
-                                for g in c["groupby_list"]]
+                                for g in c.groupby_list]
   gby_exprs = [g if isinstance(g,str) else g[0]
-                                for g in c["groupby_list"]]
+                                for g in c.groupby_list]
   comp_exprs = [compile(e,'<string>','eval') for e in gby_exprs]
   grp_table = {}
 
@@ -450,7 +418,7 @@ def processGroupByClause(c, table, prior_lcs):
     lcs = prior_lcs
     lcs.update(t.getDict())
     # Compute the key
-    k = tuple( [eval(e,globals(),lcs) for e in comp_exprs] )
+    k = tuple( [eval(e,prior_globs,lcs) for e in comp_exprs] )
     if not k in grp_table:
       grp_table[k] = []
     grp_table[k].append(t)
@@ -483,28 +451,28 @@ def processGroupByClause(c, table, prior_lcs):
 
 
 # Process where clause
-def processWhereClause(c, table, prior_lcs):
-  comp_expr = compile(c["expr"].lstrip(),"<string>","eval")
+def processWhereClause(c, table, prior_lcs, prior_globs):
+  comp_expr = compile(c.expr.lstrip(),"<string>","eval")
   for t in table:
     lcs = prior_lcs
     lcs.update(t.getDict())
-    val = eval(comp_expr, globals(), lcs)
+    val = eval(comp_expr, prior_globs, lcs)
     if val:
       yield t
 
 # Process the orderby clause
-def processOrderByClause(c, table, prior_lcs):
+def processOrderByClause(c, table, prior_lcs, prior_globs):
   # Here we do n sorts, n is the number of sort specifications
   # For each sort we first need to compute a sort value (could
   # be some expression)
 
-  sort_exprs = [ compile(os[0].lstrip(),"<string>","eval") for os in c["orderby_list"]]
-  sort_rev = [ o[1]=='desc' for o in c["orderby_list"]]
+  sort_exprs = [ compile(os[0].lstrip(),"<string>","eval") for os in c.orderby_list]
+  sort_rev = [ o[1]=='desc' for o in c.orderby_list]
 
   def computeSortSpec(tup,sort_spec):
     lcs = prior_lcs
     lcs.update(tup.getDict())
-    return eval(sort_spec, globals(), lcs)
+    return eval(sort_spec, prior_globs, lcs)
 
   sort_exprs.reverse()
   sort_rev.reverse()
@@ -550,7 +518,7 @@ def fill_in_end_vars(vars, binding_seq, i ):
 # start a new window at this location (without considering tumbling
 # windows, that check is done elsewhere).
 
-def check_start_condition(all_vars,clause,locals,var_mapping):
+def check_start_condition(all_vars,clause,locals,prior_globs,var_mapping):
   # we just need to evaluate the when expression 
   # but we need to set up the vars correctly, respecting the visibility
   # conditions
@@ -562,45 +530,45 @@ def check_start_condition(all_vars,clause,locals,var_mapping):
   locals.update( start_bindings )
 
   #evaluate the when condition
-  return eval( clause["s_when"], globals(), locals )
+  return eval( clause.s_when, prior_globs, locals )
   
 # Check the end condition of the window.
 
-def check_end_condition(vars,clause,locals,var_mapping):
+def check_end_condition(vars,clause,locals,prior_globs,var_mapping):
   # If there is no 'when' clause, return False
-  if not clause["e_when"]:
+  if not clause.e_when:
     return False
 
   end_vars = set(vars.keys()).intersection( set(var_mapping.keys()))
   end_binding = { var_mapping[v] : vars[v] for v in end_vars }
 
   locals.update( end_binding )
-  res = eval( clause["e_when"], globals(), locals)
+  res = eval( clause.e_when, prior_globs, locals)
 
   return res
 
 # Process window clause
-def processWindowClause(c, table, prior_lcs):
+def processWindowClause(c, table, prior_lcs, prior_globs):
   schema = None
   new_schema = None
 
   # Create window variable name mapping
   var_mapping = {}
-  for v in c["vars"]:
-    var_mapping[v] = c["vars"][v]
+  for v in c.vars:
+    var_mapping[v] = c.vars[v]
 		
   for t in table:
     if not schema:
       schema = t.schema
       # Create a new schema with window variables added
       new_schema = dict(t.schema)
-      for v in c["vars"]:
-        new_schema[c["vars"][v]] = len(new_schema)
+      for v in c.vars:
+        new_schema[c.vars[v]] = len(new_schema)
 
     lcs = dict(prior_lcs)
     lcs.update(t.getDict())
     # Evaluate the binding sequence
-    binding_seq = list(eval(c["in"], globals(), lcs))
+    binding_seq = list(eval(c.binding_seq, prior_globs, lcs))
 
     # Create initial window variables
 
@@ -613,10 +581,10 @@ def processWindowClause(c, table, prior_lcs):
       # Try to open a new window
       # in case of tumbling windows, only open a
       # window if there are no open windows
-      if not c["tumbling"] or (c["tumbling"] and not open_windows):
+      if not c.tumbling or (c.tumbling and not open_windows):
         vars = make_window_vars()
         fill_in_start_vars(vars,binding_seq,i)
-        if check_start_condition(vars,c,dict(lcs),var_mapping):
+        if check_start_condition(vars,c,dict(lcs),prior_globs,var_mapping):
           open_windows.append( {"window":[], "vars":vars} )
       
       new_open_windows = []
@@ -627,7 +595,7 @@ def processWindowClause(c, table, prior_lcs):
 
         fill_in_end_vars(w["vars"],binding_seq,i)
 
-        if check_end_condition(w["vars"],c,dict(lcs),var_mapping):
+        if check_end_condition(w["vars"],c,dict(lcs),prior_globs,var_mapping):
           closed_windows.append(w)
         else:
           new_open_windows.append(w)
@@ -635,7 +603,7 @@ def processWindowClause(c, table, prior_lcs):
           
     #close or remove all remaining open windows
     #if only is specified, we ignore non-closed windows
-    if not c["only"]:
+    if not c.only:
       closed_windows.extend(open_windows)
     
     # create a new tuple by extending the tuple from previous clauses
